#!/usr/bin/env python
# encoding: utf-8

# GNU GPL LICENSE
#
# This module is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; latest version thereof,
# available at: <http://www.gnu.org/licenses/gpl.txt>.
#
# This module is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this module; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA

"""
Emulates the behaviour of the BioCreative MetaServer by sending one or more
requests to an Annotation Server. This script was written during the
BioCreative II.5 competition and therefor is currently limited to full-text
query functionality only.

The script request annotations for a given full-text file (either utf8 or
xml) form the Annotation Server, evaluates the results are in the correct
format and then optionally writes the result(s) to a tab-separation
formated file that can be used as input for the BC evaluation script.

Written for python2.5 for broader usability, but works in 2.6 just as well if
you don't mind the deprecation warinings.

Created by Florian Leitner on 2009-03-18 at CNIO.
Copyright (c) 2009 Florian Leitner. All rights reserved.
"""

import logging
import os
import re
import socket
import sys

from optparse import OptionParser
from xml.parsers.expat import ExpatError
from xmlrpclib import Fault, ProtocolError

from bcms_client_lib.evaluation import evaluate_result as evaluate
from bcms_client_lib.io_tools import load_article, print_results
from bcms_client_lib.xmlrpc import ServerClient

__author__ = "Florian Leitner"
__version__ = "2.0"

def run(server_uri, timeout, input_files, output_handles=None):
    file_suffix = re.compile('(.noSDA)?(.xml)?(.utf8)?$')
    client = ServerClient(server_uri, timeout)
    logger = logging.getLogger("Request")
    success = 0
    
    for filename in input_files:
        # ====================
        # = Load the article =
        # ====================
        
        if not os.access(filename, os.R_OK):
            # use logging, this is a root logger problem
            logging.getLogger(os.path.basename(sys.argv[0])).critical(
                "can't read input file '%s'" % filename
            )
            return -1 # no point in continuing
        
        if not any(file_suffix.search(filename).groups()):
            # use logging, this is a root logger problem
            logging.getLogger(os.path.basename(sys.argv[0])).critical(
                "input file '%s' has the wrong suffix" % filename
            )
            return -1 # no point in continuing
        
        uid = file_suffix.sub(
            "", os.path.basename(filename)
        ).replace("_", "/")
        
        try:
            logger.debug("loading text for %s" % uid)
            article = load_article(filename)
        except IOError:
            logger.critical("I/O error while reading '%s'" % filename)
            return -1 # no point in continuing
        
        # ====================
        # = Make the request =
        # ====================
        
        try:
            logger.debug("requesting results for %s" % uid)
            result = client.fulltext_request(article)
        except socket.timeout, e:
            logger.error(
                "configured timeout reached (%i s)." % timeout
            )
            success = -1
        except socket.error, e:
            logger.critical(
                "socket error, server not reachable: %s (%s)." % (
                    # unicode(e.errmsg), unicode(e.errno) # py2.6
                    unicode(e.args[1]), unicode(e.args[0]) # py2.5
                )
            )
            return -1 # no point in continuing
        except ProtocolError, e:
            logger.critical(
                "HTTP error, e.g., proxy server failure: %s (%s)." % (
                    # unicode(e.errmsg), unicode(e.errcode) # py2.6
                    unicode(e.args[1]), unicode(e.args[0]) # py2.5
                )
            )
            return -1 # no point in continuing
        except Fault, e:
            logger.error(
                "fault response received: %s (%s)." % (
                    unicode(e.faultString), unicode(e.faultCode)
                )
            )
            success = -1
        except ExpatError, e:
            logger.critical(
                u"XML parser error, malformed XML data: %s." % unicode(e)
            )
            return -1 # no point in continuing
        except Exception, e:
            logger.critical("Unpredicted exception detected.")
            logger.exception("Exception was:")
            return -1 # no point in continuing
        else:
            # ===========================================
            # = Evaluate and possibly write the results =
            # ===========================================
            
            logger.debug(u"evaluating results for %s" % uid)
            
            if evaluate(result):
                if output_handles is not None:
                    logger.debug(u"writing results for %s" % uid)
                    print_results(uid, result, output_handles)
            else:
                logger.warn(u"results for %s were invalid" % uid)
                
                if success == 0:
                    success = 1
    
    return success

def main():
    usage = "usage: %prog [-h --help --options] <article[s...]>"
    description = \
        "request results from an Annotation Server and assert data " \
        "integrity of any received BCMS result types (interaction, " \
        "mappings, mentions, normalizations, pairs, taxons), " \
        "optionally printing BC II.5-relevant data types to files"
    epilog = \
        "BC II.5 result files are only created if a file name for the " \
        "corresponding result type is given as an optional argument " \
        "('--ACT <filename>', '--INT <filename>', or '--IPT <filename>'). " \
        "This script exits with " \
        "-1 when encountering technical errors, " \
        "1 if a data integrity test failed " \
        "and 0 only if all was OK (in that order). " \
        "Structured UTF-8 text input files are assumed to have the " \
        "'.utf8' suffix, creating the document ID from the file name. " \
        "It sends plain XML if '.xml' suffix files are used as input. " \
        "Note that it can be run with Python 2.5 and 2.6, " \
        "but using this script on OSX requires the use of the native " \
        "Python version Apple supplies with its operating system " \
        "due to the faulty implementation of the poll I/O multiplexer by " \
        "Apple (see 'man poll', section BUGS, on OSX). " \
        "Using it on Windows and any *nix should work properly " \
        "with both versions of Python."
    parser = OptionParser(
        usage=usage, version=__version__, prog=os.path.basename(sys.argv[0]),
        description=description, epilog=epilog, add_help_option=True
    )
    
    # ==============================================================
    # = option parser options and default settings for this script =
    # ==============================================================
    
    parser.set_defaults(log_level=logging.WARNING)
    parser.add_option(
        "-s", "--server", action="store", type="string",
        default="http://localhost:8000",
        help=u"server URL of the BCAS [default: %default]"
    )
    parser.add_option(
        "-t", "--timeout", action="store", type="int", default=600,
        help=u"server address of the BCAS [default: %default]"
    )
    parser.add_option(
        "-d", "--dir", action="store", type="string", default=os.getcwd(),
        help=u"base directory for any BC II.5 task output file [default: %default]"
    )
    parser.add_option(
        "--ACT", action="store", type="string", default=None,
        help=u"ACT/interaction output file (if any) [default: %default]"
    )
    parser.add_option(
        "--INT", action="store", type="string", default=None,
        help=u"INT/normalizations output file (if any) [default: %default]"
    )
    parser.add_option(
        "--IPT", action="store", type="string", default=None,
        help=u"IPT/pairs output file (if any) [default: %default]"
    )
    
    for log_level in ("debug", "info", "warning", "error", "critical"):
        parser.add_option(
            "--%s" % log_level, action="store_const", dest="log_level",
            const=getattr(logging, log_level.upper()),
            help=u"set log level to %s [default: %s]" % (
                log_level, "warning"
            )
        )
    
    parser.add_option(
        "--quiet", action="store_const", dest="log_level",   
        const=logging.CRITICAL + 10,
        help=u"log no events at all [default: %s]" % "warning"
    )
    
    opts, args = parser.parse_args()
    logging.basicConfig(
        level=opts.log_level,
        format='%(asctime)s %(levelname)-8s %(name)s: %(message)s'
    )
    logger = logging.getLogger(os.path.basename(sys.argv[0]))
    logger.debug("processing options and arguments")
    
    if len(args) == 0:
        parser.error("no input files")
    
    # =================================================
    # = OS-independent, safe creation of output files =
    # =================================================
    
    if not os.path.exists(opts.dir):
        logger.critical("path '%s' doesn't seem to exist" % opts.dir)
        return -1
    
    if not os.path.isdir(opts.dir):
        logger.critical("path '%s' not a directory" % opts.dir)
        return -1
    
    output_handles = dict()
    
    for kind in ("ACT", "INT", "IPT"):
        filename = getattr(opts, kind)
        
        if filename is not None:
            if not os.access(opts.dir, os.W_OK):
                logger.critical("can't write to directory '%s'" % opts.dir)
                return -1
            
            full_path = os.path.normpath(os.path.join(opts.dir, filename))
            
            try:
                output_handles[kind] = open(full_path, 'w')
                logger.debug("%s results will be written to '%s'" % (
                    kind, full_path
                ))
            except IOError, e:
                logger.error(
                    "I/O error while opening '%s': %s - not creating "
                    "%s output" % (full_path, unicode(e), kind)
                )
    
    if not len(output_handles):
        output_handles = None
    else:
        logger.info("results for %s will be stored" % ", ".join(
            output_handles.keys()
        ))
    
    # ============================================
    # = Run the script with the given parameters =
    # ============================================
    
    logger.info("starting %i annotation requests" % len(args))
    success = run(opts.server, opts.timeout, args, output_handles)
    
    # ===================================
    # = Close the output files (if any) =
    # ===================================
    
    if output_handles is not None:
        logger.debug("closing output files")
        
        for kind, filehandle in output_handles.items():
            try:
                filehandle.close()
            except IOError, e:
                logger.error(
                    "I/O error while closing %s file: %s" % (kind, e.message)
                )
    
    logger.info("exiting with status %i" % success)
    return success

if __name__ == "__main__":
    sys.exit(main())
