# encoding: utf-8

# GNU GPL LICENSE
#
# This module is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; latest version thereof,
# available at: <http://www.gnu.org/licenses/gpl.txt>.
#
# This module is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this module; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA

"""
The BCMS test client is a reimplementation of some of the rudimentary result
assertions the BCMS makes on received results, and can print BC II.5 relevant
results to files.

This script is intended to asses the integrity of an Annotation Server without having to cope with the connectivity problems of the Internet. If running it
with an Annotation Server does not exit cleanly, the server will not work
properly with the BCMS.

Created by Florian Leitner on 2009-05-10 at CNIO.
Copyright (c) 2009 Florian Leitner. All rights reserved.
"""

import logging

from functools import wraps

__author__ = "Florian Leitner"
__version__ = "1.0"

_ALLOWED_DATABASES = (
    "EntrezGene",
    "GenBank",
    "GenPept",
    "NCBI_Tax",
    "OMIM",
    "PDB",
    "PIR",
    "PubChemSubstance",
    "RefSeq",
    "UniGene",
    "UniProt",
)

# =======================
# = Per Item Assertions =
# =======================

# ASSERTIONS, WRAPPERS, HELPERS #

def _assert_type(item, key, type_):
    """Assert field value has the correct type."""
    if item.has_key(key):
        value = item[key]
        if not isinstance(value, type_):
            try:
                item[key] = type_(value)
            except ValueError, TypeError:
                raise AssertionError("%s must be of type %s" % (
                    key, type_.__name__
                ))
    

def _assert_required(item, *keys):
    """Assert the given fields exist."""
    for key in keys:
        if not item.has_key(key):
            raise AssertionError("%s is a required key" % key)
    

def _assert_database(name):
    """Assert the database is in the list of known databases."""
    if name not in _ALLOWED_DATABASES:
        raise AssertionError("database name '%s' is illegal" % name)

def annotation(fun):
    """Base wrapper for any annotation item."""
    
    @wraps(fun)
    def wrapper(item):
        """
        Assert confidence is given as float and is in range ]0..]
        and the fields version and evidence have the correct type if given.
        """
        _assert_required(item, 'confidence')
        
        fun(item)
        
        for key, klass in (
                ('version', basestring),
                ('confidence', float),
                ('evidence', basestring),
            ):
            _assert_type(item, key, klass)
        
        if not 0.0 < item['confidence'] <= 1.0:
            raise AssertionError("confidence must be in ]0..1] range")
    
    return wrapper

def ranked_annotation(fun):
    """Extended base wrapper for any ranked annotation item."""
    
    @wraps(fun)
    @annotation
    def wrapper(item):
        """Assert rank is given as integer."""
        _assert_required(item, 'rank')
        fun(item)
        _assert_type(item, 'rank', int)
    
    return wrapper

def _assert_text_mapping(item):
    """Assert all fields to map to a string in a Text object are given."""
    _assert_required(item, 'mention', 'offset', 'section')
    
    for key, klass in (
            ('mention', basestring),
            ('offset', int),
            ('section', basestring),
            ('qualifier', basestring),
            ('number', int),
        ):
        _assert_type(item, key, klass)
    

def _assert_identifier(item):
    """Assert all fields to define an Identifier are given."""
    _assert_required(item, 'accession')
    
    if item.has_key('database'):
        _assert_database(item['database'])
    
    for key, klass in (
            ('database', basestring),
            ('accession', basestring),
        ):
        _assert_type(item, key, klass)
    

def _pair_helper(item):
    """Helper function for Identifier pairs."""
    for db in ('database_a', 'database_b'):
        if item.has_key(db):
            _assert_database(item[db])
    
    if item.has_key('is_auto') and item['is_auto']:
        test = dict()
        
        for db in ('database_a', 'database_b'):
            test[db] = item[db] if item.has_key(db) else 'UniProt'
        
        if test['database_a'] != test['database_b']:
            raise AssertionError("is auto, but databases not equal")
        
        if item['accession_a'] != item['accession_b']:
            raise AssertionError("is auto, but accessions not equal")
    
    for key, klass in (
            ('database_a', basestring),
            ('accession_a', basestring),
            ('database_b', basestring),
            ('accession_b', basestring),
            ('is_auto', bool),
        ):
        _assert_type(item, key, klass)
    

# SINGLE ANOTATION EVALUATION #

@annotation
def interaction(item):
    _assert_required(item, 'has_interaction')
    _assert_type(item, 'has_interaction', bool)

@ranked_annotation
def mapping(item):
    _assert_text_mapping(item)
    _assert_identifier(item)

@ranked_annotation
def mention(item):
    _assert_text_mapping(item)

@ranked_annotation
def normalization(item):
    _assert_identifier(item)
    _assert_type(item, 'is_interactor', bool)

@ranked_annotation
def pair(item):
    _assert_required(item, 'accession_a', 'accession_b')
    _pair_helper(item)

@ranked_annotation
def taxon(item):
    """taxons"""
    _assert_required(item, 'accession')
    _assert_type(item, 'accession', basestring)

# =======================================
# = Complete Result Type Set Assertions =
# =======================================

# ASSERTIONS, HELPERS, WRAPPERS #

def group(assert_fun):
    """Base decorator with the single item assertion function as argument."""
    
    def decorator(fun):
        """Base wrapper for result type sets."""
        
        @wraps(fun)
        def wrapper(group):
            """
            If a set is given, assert it is a list, execute the per item
            assertions and call the group functions.
            """
            if group is None:
                return
                
            if not isinstance(group, list):
                raise AssertionError(
                    "result type items must be in a list"
                )
            
            for item in group:
                assert_fun(item)
            
            fun(group)
        
        return wrapper
    
    return decorator

def ranked_group(assert_fun):
    """Extended base decorator for ranked result type sets."""
    
    def decorator(fun):
        """Extended wrapper for ranked result type sets."""
        
        @wraps(fun)
        @group(assert_fun)
        def wrapper(group):
            """
            Assert the ranks given are unique and resemble the range(1, n).
            """
            for item in group:
                assert_fun(item)
            
            ranks = [i['rank'] for i in group]
            ranks.sort()
            
            if ranks != range(1, len(group) + 1):
                raise AssertionError("ranks not a consecutive list from 1")
            
            fun(group)
        
        return wrapper
    
    return decorator

def _assert_unique(group, *keys):
    """
    Assert no other item in the group has the same values for the given
    keys.
    """
    select = (lambda item_, keys_: set([
        item_[k] for k in keys_ if k in item_
    ]))
    items = [select(item, keys) for item in group]
    
    while len(items):
        current = items.pop()
        
        if current in items:
            raise AssertionError("non-unique result set")
    

def _add_uniprot_database(group, *keys):
    group_copy = group
    changes = 0
    
    for idx, item in enumerate(group):
        item_copy = item
        changed = False
        
        for key in keys:
            if not item.has_key(key):
                changes += 1
                
                if changes == 1:
                    group_copy = list(group_copy)
                
                if not changed:
                    item_copy = item_copy.copy()
                
                item_copy[key] = 'UniProt'
                changed = True
        
        if changed:
            group_copy[idx] = item_copy
    
    return group_copy

def _islast(sequence):
    """Generate (islast, item) pairs."""
    iterable = iter(sequence)
    previous = iterable.next()
    
    for item in iterable:
        yield False, previous
        previous = item
    
    yield True, previous

def format(group):
    """
    Yield lines for the group's items formated as representational strings.
    """
    for last, item in _islast(group):
        keys = item.keys()
        keys.sort()
        
        for key in keys:
            yield "%-15s: %s" % (key, repr(item[key]))
        
        if not last:
            yield "----------------"
    

# RESULT SET EVALUATION #

@group(interaction)
def interactions(group): # ACT
    """interaction""" # result set key
    if len(group) > 1:
        raise AssertionError("only one interaction allowed")

@ranked_group(mapping)
def mappings(group):
    """mappings""" # result set key
    databased_group = _add_uniprot_database(group, 'database')
    _assert_unique(
        databased_group, 
        'mention', 'offset', # mention
        'section', 'qualifier', 'number', # Text
        'database', 'accession' # Identifier
    )

@ranked_group(mention)
def mentions(group):
    """mentions""" # result set key
    _assert_unique(
        group,
        'mention', 'offset', # mention
        'section', 'qualifier', 'number', # Text
    )

@ranked_group(normalization)
def normalizations(group):
    """normalizations""" # result set key
    databased_group = _add_uniprot_database(group, 'database')
    _assert_unique(
        databased_group,
        'database', 'accession' # Identifier
    )

@ranked_group(pair)
def pairs(group):
    """pairs""" # result set key
    databased_group = _add_uniprot_database(group, 'database_a', 'database_b')
    _assert_unique(
        databased_group,
        'database_a', 'accession_a', # Identifier A
        'database_b', 'accession_b' # Identifier B
    )

@ranked_group(taxon)
def taxons(group):
    """taxons""" # result set key
    _assert_unique(
        group,
        'accession' # Identifier (NCBI_Tax)
    )

# ============================
# = Main Evaluation Function =
# ============================

def evaluate_result(data):
    """Evaluate all result sets are in the correct format."""
    evaluate = dict()
    logger = logging.getLogger("Evaluation")
    valid = True
    
    for evaluation_fun in (
        interactions, mappings, mentions, normalizations, pairs, taxons
        ):
        evaluate[evaluation_fun.__doc__] = evaluation_fun
    
    for result_type, result_set in data.items():
        try:
            evaluate[result_type](result_set)
        except AssertionError, e:
            valid = False
            logger.warn("%s: %s" % (result_type, e.message))
            
            if logger.isEnabledFor(logging.INFO):
                try:
                    for line in format(result_set):
                        logger.info(line)
                except Exception:
                    logger.info("result data malformed: %s" % str(result_set))
        except KeyError:
            valid = False
            logger.warn("unknown result type set '%s'" % result_type)
    
    return valid
