# encoding: utf-8

# GNU GPL LICENSE
#
# This module is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; latest version thereof,
# available at: <http://www.gnu.org/licenses/gpl.txt>.
#
# This module is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this module; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA

"""
I/O library for the test client.

Created by Florian Leitner on 2009-05-10 at CNIO.
Copyright (c) 2009 Florian Leitner. All rights reserved.
"""

import logging

from itertools import count

__author__ = "Florian Leitner"
__version__ = "1.0"

_bc25_logger = logging.getLogger("BC II.5")

# =========
# = INPUT =
# =========

def _load_xml(filehandle):
    return filehandle.read().decode('utf-8').strip()

def _load_utf8(filehandle):
    # TODO: real, stateful parsing...
    QUALIFIER = "QUALIFIER "
    KEY = "KEY "
    content = list()
    last = dict()
    key = None
    section = None
    before = None
    number = None
    
    for line in filehandle:
        nice = line.decode('utf-8').strip()
        ll = len(nice)
        
        if ll:
            if ll >= 4 and nice[:4] == KEY:
                if len(nice[4:].strip()):
                    key = nice[4:].strip()
            elif ll >= 10 and line[:10] == QUALIFIER:
                if len(line[10:].strip()):
                    if key is None:
                        last['qualifier'] = nice[10:].strip()
                    else:
                        last['qualifier'] = u"%s: %s" % (
                            nice[10:].strip(), key
                        )
            else:
                split = nice.index(' ')
                section = nice[:split].lower()
                last['section'] = section
                last['content'] = nice[split + 1:]
        elif len(last):
            if last.has_key('qualifier'):
                if last['qualifier'] == before:
                    if number is None:
                        number = count()
                        content[-1]["number"] = number.next()
                    
                    last["number"] = number.next()
                else:
                    number = None
                
                before = last['qualifier']
            
            content.append(last)
            last = dict()
            key = None
            section = None
    
    return content

def load_article(filename):
    """Return the content of the article in its RPC parameterized form."""
    # readability of the file is assumed to be checked by the caller
    load = _load_xml if filename[-3:].lower() == "xml" else _load_utf8
    article = open(filename, 'r')
    
    try:
        content = load(article)
    except IOError, e:
        logging.getLogger("Filereader").error("I/O error on %s: %s" (
            filename, e.message
        ))
    article.close()
    return content

# ==========
# = OUTPUT =
# ==========

def _skip(result_type, item, *db_keys):
    for db_key in db_keys:
        if item.has_key(db_key) and item[db_key] != 'UniProt':
            _bc25_logger.info(
                "%s result for non-UniProt database %s skipped" % (
                    result_type, item[db_key]
                )
            )
            return True
    
    return False

def interaction(uid, item):
    """ACT"""
    return u"%s\t%s\t%.6f" % (
        uid, item['has_interaction'], item['confidence']
    )

def normalizations(uid, item):
    """INT"""
    if _skip(normalizations.__doc__, item, 'database'):
        return None
    
    return u"%s\t%s\t%i\t%.6f" % (
        uid, item['accession'], item['rank'], item['confidence']
    )

def pairs(uid, item):
    """IPT"""
    if _skip(pairs.__doc__, item, 'database_a', 'database_b'):
        return None
    
    return u"%s\t%s\t%s\t%i\t%.6f" % (
        uid, item['accession_a'], item['accession_b'],
        item['rank'], item['confidence']
    )

def print_results(uid, results, output_handles):
    for format in (interaction, normalizations, pairs):
        if format.__name__ in results and format.__doc__ in output_handles:
            for item in results[format.__name__]:
                line = format(uid, item)
                
                if line is not None:
                    print >> output_handles[format.__doc__], line
        else:
            _bc25_logger.debug("no %s data in result (has: %s)" % (
                format.__doc__, ", ".join(results.keys())
            ))
    

