package biocreative.webservice;

import java.util.ArrayList;
import java.util.List;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

/**Data structure for the global result Map.
 *
 * Example of how to create a global Result HashMap, for example in the
 * getAnnotation() methods of either Fulltext or Medline after your text
 * mining pipeline has calculated the results:
 *
 * Result result = new Result();
 * result.addInteraction(has_interaction, interaction_confidence);
 * result.addPair(accession_a, accession_b, rank, confidence);
 * result.addNormalization(accession_a, rank, confidence);
 * result.addNormalization(accession_b, rank, confidence);
 * return result;
 */
public class Result extends HashMap {
    
    /**Return a HashMap filled with the content array, where keys are the
     * even and values the odd elements.
     */
    private static Map fillMap(Object[] content) {
        HashMap map = new HashMap();
        for (int i = 0; i < content.length; i+=2) {
            map.put(content[i], content[i+1]);
        }
        return map;
    }
    
    /**Put a single result HashMap to the global results. */
    private Map putResult(String key, Object[] content) {
        Map result = Result.fillMap(content);
        if (!this.containsKey(key)) this.put(key, new ArrayList());
        ((List) this.get(key)).add(result);
        return result;
    }
    
    /**Add the interaction classification result to the global results. */
    public Map addInteraction(boolean has_interaction, double confidence) {
        Map interaction = Result.fillMap(new Object[] {
            "has_interaction", has_interaction,
            "confidence", confidence
        });
        this.put("interaction", new Object[] {interaction});
        return interaction;
    }
    
    /**Add a mapping result to the global results.
     *
     * The BCMS will use defaults for the missing pairs:
     * qualifier => "" (none)
     * number => null (none)
     * database => "UniProt"
     * is_interactor => false
     */
    public Map addMapping(
        String mention, int offset,
        String section,
        int rank, double confidence,
        String accession
    ) {
        return this.putResult("mappings", new Object[] {
            "mention", mention,
            "offset", offset,
            "section", section,
            "accession", accession,
            "rank", rank,
            "confidence", confidence
        });
    }
    
    /**Add a mapping result to the global results.
     *
     * The BCMS will use defaults for the missing pairs:
     * qualifier => "" (none)
     * number => null (none)
     * is_interactor => false
     */
    public Map addMapping(
        String mention, int offset,
        String section,
        int rank, double confidence,
        String database, String accession
    ) {
        return this.putResult("mappings", new Object[] {
            "mention", mention,
            "offset", offset,
            "section", section,
            "database", database,
            "accession", accession,
            "rank", rank,
            "confidence", confidence
        });
    }
    
    /**Add a mapping result to the global results.
     *
     * The BCMS will use defaults for the missing pairs:
     * qualifier => "" (none)
     * number => null (none)
     * database => "UniProt"
     */
    public Map addMapping(
        String mention, int offset,
        String section,
        int rank, double confidence,
        String accession, boolean is_interactor
    ) {
        return this.putResult("mappings", new Object[] {
            "mention", mention,
            "offset", offset,
            "section", section,
            "accession", accession,
            "is_interactor", is_interactor,
            "rank", rank,
            "confidence", confidence
        });
    }
    
    /**Add a mapping result to the global results.
     *
     * The BCMS will use defaults for the missing pairs:
     * qualifier => "" (none)
     * number => null (none)
     */
    public Map addMapping(
        String mention, int offset,
        String section,
        int rank, double confidence,
        String database, String accession, boolean is_interactor
    ) {
        return this.putResult("mappings", new Object[] {
            "mention", mention,
            "offset", offset,
            "section", section,
            "database", database,
            "accession", accession,
            "is_interactor", is_interactor,
            "rank", rank,
            "confidence", confidence
        });
    }
    
    /**Add a mapping result to the global results.
     *
     * The BCMS will use defaults for the missing pairs:
     * number => null (none)
     * database => "UniProt"
     * is_interactor => false
     */
    public Map addMapping(
        String mention, int offset,
        String section, String qualifier,
        int rank, double confidence,
        String accession
    ) {
        return this.putResult("mappings", new Object[] {
            "mention", mention,
            "offset", offset,
            "section", section,
            "qualifier", qualifier,
            "accession", accession,
            "rank", rank,
            "confidence", confidence
        });
    }
    
    /**Add a mapping result to the global results.
     *
     * The BCMS will use defaults for the missing pairs:
     * number => null (none)
     * is_interactor => false
     */
    public Map addMapping(
        String mention, int offset,
        String section, String qualifier,
        int rank, double confidence,
        String database, String accession
    ) {
        return this.putResult("mappings", new Object[] {
            "mention", mention,
            "offset", offset,
            "section", section,
            "qualifier", qualifier,
            "database", database,
            "accession", accession,
            "rank", rank,
            "confidence", confidence
        });
    }
    
    /**Add a mapping result to the global results.
     *
     * The BCMS will use defaults for the missing pairs:
     * number => null (none)
     * database => "UniProt"
     */
    public Map addMapping(
        String mention, int offset,
        String section, String qualifier,
        int rank, double confidence,
        String accession, boolean is_interactor
    ) {
        return this.putResult("mappings", new Object[] {
            "mention", mention,
            "offset", offset,
            "section", section,
            "qualifier", qualifier,
            "accession", accession,
            "is_interactor", is_interactor,
            "rank", rank,
            "confidence", confidence
        });
    }
    
    /**Add a mapping result to the global results.
     *
     * The BCMS will use defaults for the missing pairs:
     * number => null (none)
     */
    public Map addMapping(
        String mention, int offset,
        String section, String qualifier,
        int rank, double confidence,
        String database, String accession, boolean is_interactor
    ) {
        return this.putResult("mappings", new Object[] {
            "mention", mention,
            "offset", offset,
            "section", section,
            "qualifier", qualifier,
            "database", database,
            "accession", accession,
            "is_interactor", is_interactor,
            "rank", rank,
            "confidence", confidence
        });
    }
    
    /**Add a mapping result to the global results.
     *
     * The BCMS will use defaults for the missing pairs:
     * database => "UniProt"
     */
    public Map addMapping(
        String mention, int offset,
        String section, String qualifier, int number,
        int rank, double confidence,
        String accession, boolean is_interactor
    ) {
        return this.putResult("mappings", new Object[] {
            "mention", mention,
            "offset", offset,
            "section", section,
            "qualifier", qualifier,
            "number", number,
            "accession", accession,
            "is_interactor", is_interactor,
            "rank", rank,
            "confidence", confidence
        });
    }
    
    /**Add a mapping result to the global results. */
    public Map addMapping(
        String mention, int offset,
        String section, String qualifier, int number,
        int rank, double confidence,
        String database, String accession, boolean is_interactor
    ) {
        return this.putResult("mappings", new Object[] {
            "mention", mention,
            "offset", offset,
            "section", section,
            "qualifier", qualifier,
            "number", number,
            "database", database,
            "accession", accession,
            "is_interactor", is_interactor,
            "rank", rank,
            "confidence", confidence
        });
    }
    
    /**Add a mention result to the global results.
     *
     * The BCMS will use defaults for the missing pairs:
     * qualifier => "" (none)
     * number => null (none)
     */
    public Map addMention(
        String mention, String offset,
        String section,
        int rank, double confidence
    ) {
        return this.putResult("mentions", new Object[] {
            "mention", mention,
            "offset", offset,
            "section", section,
            "rank", rank,
            "confidence", confidence
        });
    }
    
    /**Add a mention result to the global results.
     *
     * The BCMS will use defaults for the missing pairs:
     * number => null (none)
     */
    public Map addMention(
        String mention, String offset,
        String section, String qualifier,
        int rank, double confidence
    ) {
        return this.putResult("mentions", new Object[] {
            "mention", mention,
            "offset", offset,
            "section", section,
            "qualifier", qualifier,
            "rank", rank,
            "confidence", confidence
        });
    }
    
    /**Add a mention result to the global results. */
    public Map addMention(
        String mention, String offset,
        String section, String qualifier, int number,
        int rank, double confidence
    ) {
        return this.putResult("mentions", new Object[] {
            "mention", mention,
            "offset", offset,
            "section", section,
            "qualifier", qualifier,
            "number", number,
            "rank", rank,
            "confidence", confidence
        });
    }
    
    /**Add a normalization result to the global results.
     *
     * The BCMS will use default for the missing pairs:
     * database => "UniProt"
     * is_interactor => false (ignored during the BC II.5 challenge!)
     */
    public Map addNormalization(
        String accession, int rank, double confidence
    ) {
        return this.putResult("normalizations", new Object[] {
            "accession", accession,
            "rank", rank,
            "confidence", confidence
        });
    }
    
    /**Add a normalization result to the global results.
     *
     * The BCMS will use default for the missing pairs:
     * database => "UniProt"
     */
    public Map addNormalization(
        String accession, boolean is_interactor, int rank, double confidence
    ) {
        return this.putResult("normalizations", new Object[] {
            "accession", accession,
            "is_interactor", is_interactor,
            "rank", rank,
            "confidence", confidence
        });
    }
    
    /**Add a normalization result to the global results.
     *
     * The BCMS will use default for the missing pairs:
     * is_interactor => false (ignored during the BC II.5 challenge!)
     */
    public Map addNormalization(
        String database, String accession, int rank, double confidence
    ) {
        return this.putResult("normalizations", new Object[] {
            "accession", accession,
            "database", database,
            "rank", rank,
            "confidence", confidence
        });
    }
    
    /**Add a normalization result to the global results. */
    public Map addNormalization(
        String database, String accession,
        boolean is_interactor, int rank, double confidence
    ) {
        return this.putResult("normalizations", new Object[] {
            "accession", accession,
            "database", database,
            "is_interactor", is_interactor,
            "rank", rank,
            "confidence", confidence
        });
    }
    
    /**Add a pair result to the global results.
     *
     * The BCMS will use defaults for the missing pairs:
     * database_X => "UniPort"
     * is_auto => false (ignored during the BC II.5 challenge!)
     */
    public Map addPair(
        String accession_a, String accession_b, int rank, double confidence
    ) {
        return this.putResult("pairs", new Object[] {
            "accession_a", accession_a,
            "accession_b", accession_b,
            "rank", rank,
            "confidence", confidence            
        });
    }
    
    /**Add a pair result to the global results.
     *
     * The BCMS will use defaults for the missing pairs:
     * database_X => "UniPort"
     */
    public Map addPair(
        String accession_a, String accession_b,
        boolean is_auto, int rank, double confidence
    ) {
        return this.putResult("pairs", new Object[] {
            "accession_a", accession_a,
            "accession_b", accession_b,
            "is_auto", is_auto,
            "rank", rank,
            "confidence", confidence            
        });
    }
    
    /**Add a pair result to the global results.
     *
     * The BCMS will use defaults for the missing pairs:
     * is_auto => false (ignored during the BC II.5 challenge!)
     */
    public Map addPair(
        String database_a, String accession_a,
        String database_b, String accession_b,
        int rank, double confidence
    ) {
        return this.putResult("pairs", new Object[] {
            "database_a", database_a,
            "accession_a", accession_a,
            "database_b", database_b,
            "accession_b", accession_b,
            "rank", rank,
            "confidence", confidence            
        });
    }
    
    /**Add a pair result to the global results. */
    public Map addPair(
        String database_a, String accession_a,
        String database_b, String accession_b,
        boolean is_auto, int rank, double confidence
    ) {
        return this.putResult("pairs", new Object[] {
            "database_a", database_a,
            "accession_a", accession_a,
            "database_b", database_b,
            "accession_b", accession_b,
            "is_auto", is_auto,
            "rank", rank,
            "confidence", confidence            
        });
    }
    
    /**Add a taxon result to the global results. */
    public Map addTaxon(
        String accession, int rank, double confidence
    ) {
        return this.putResult("taxa", new Object[] {
            "accession", accession,
            "rank", rank,
            "confidence", confidence
        });
    }
    
    /**Sample method to populate the global results with some hypothetical
     * results.
     *
     * This is only for testing purposes and should not be used in production!
     */
    public Map sampleResult() {
        this.addInteraction(false, 0.34);
        this.addPair("P1234", "Q1234", 1, 1.0);
        this.addPair("P1235", "Q1235", 2, 0.5);
        this.addPair("P1236", "Q1236", 3, 0.0001);
        this.addNormalization("P1234", 1, 1.0);
        this.addNormalization("P1235", 2, 0.5);
        this.addNormalization("P1236", 3, 0.0001);
        return this;
    }
}