#!/usr/bin/env perl

use strict;
use utf8;
use Getopt::Long;
use Pod::Usage;
use RPC::XML;
use RPC::XML::Server;

=head1 NAME

BioCreative Annotation Server (BCAS)

=head1 AUTHOR

Florian Leitner

=head1 VERSION

1.0

=head1 SYNOPSIS

bc-annotation-server.pl [options]

=head1 OPTIONS

=over 8

=item B<-help>
Print a brief help message and exit.

=item B<-host>
Hostname/IP address (default: localhost).

=item B<-port>
Portnumber (default: 8000).

=item B<-xml>
Server should use XML instead of structured UTF-8 as input.

=back

=head1 DESCRIPTION

B<This program> is an example of how to implement the BCAS server. The actual
text mining API needs to be implemented in the subs get_fulltext_annotations,
get_medline_annotations, and/or get_xml_annotations, but
should either be finetuned by your needs or exchanged for your own subs.

Fulltext handles annotation request for full-text articles (as in the BC II.5
challenge), while Medline handles annotation requests for PubMed abstracts (as
in the BioCreative MetaServer prototype) using the PMID as parameters. For the
challenge, only the full-text implementation is necessary.

The XML-RPC API is very simple:
<Type>.getAnnotation
where Type is either 'Fulltext' or 'Medline'.

The client will call this server using either fulltext or the PMID as input
and expects the Annotation Server to return a struct (dict) of result types
("interaction", "normalizations", "pairs", plus three more for the full API
not used in the contest), where each value is an array of structs with the
specific result. These result structs always contain a "confidence" and "rank"
key, except for the array of length one in "interction", where it makes no
sense to report a rank. For details, see the description on the team webpage.

=head1 FUNCTIONS

=head2 sample_result

Sample result structure for an article. For initial testing purposes only
and shuld be removed once the pipeline is connected to the server.

=cut

sub sample_result {
  return {
    "pairs" => [
      {
        "accession_a" => "P1234", "accession_b" => "Q1234",
        "rank" => 1, "confidence" => 1.0
      },
      {
        "accession_a" => "P1235", "accession_b" => "Q1235",
        "rank" => 2, "confidence" => 0.5
      },
      {
        "accession_a" => "P1236", "accession_b" => "Q1236",
        "rank" => 3, "confidence" => 0.0001
      },
    ],
    "interaction" => [
      {"has_interaction" => 0, "confidence" => 0.34},
    ],
    "normalizations" => [
      {"accession" => "P1234", "rank" => 1, "confidence" => 1.0},
      {"accession" => "P1235", "rank" => 2, "confidence" => 0.5},
      {"accession" => "P1236", "rank" => 3, "confidence" => 0.0001},
    ],
  }
}

=head2 get_fulltext_annotation

API function - edit this function or replace it

=cut

sub get_fulltext_annotation {
  my $self = shift;
  my $array_ref = shift;
  
  foreach my $h (@$array_ref) {
    if (ref $h eq "HASH") {
      foreach my $k (keys %$h) {
        $h->{$k} = utf8::encode($h->{$k}) unless (utf8::is_utf8($h->{$k}));
      }
    } else {
      return RPC::XML::fault->new(
        3, 'illegal type ('.(ref $h).') for text array element'
      );
    }
  }
  
  # ==================== #
  # INSERT API CALL HERE #
  # ==================== #
  return sample_result();
}

=head2 get_fulltext_annotation

API function - edit this function or replace it

=cut

sub get_medline_annotation {
  my $self = shift;
  my $pmid = shift;
  
  # here you could check if the PMID exists in your DB
  
  # ==================== #
  # INSERT API CALL HERE #
  # ==================== #
  return sample_result();  
}

=head2 get_fulltext_annotation

API function - edit this function or replace it

=cut


sub get_xml_annotation {
  my $self = shift;
  my $text = shift;
  
  $text = utf8::encode($text) unless (utf8::is_utf8($text));
  
  # ==================== #
  # INSERT API CALL HERE #
  # ==================== #
  return sample_result();
}

=head2 get_fulltext_annotation

parse command line options and/or set your default server values here

=cut

sub getopt {
  my $help = 0;
  my $host = 'localhost';
  my $port = 8000;
  my $xml = 0;
  
  GetOptions(
    'help|?' => \$help,
    'port=i' => \$port,
    'host=s' => \$host,
    'xml' => \$xml
  ) or pod2usage(2);
  
  pod2usage(1) if $help;
  
  return $host, $port, $xml;
}

# ==================== #
# start/run the server #
# ==================== #

my ($host, $port, $xml) = getopt();
my $server = RPC::XML::Server->new(
  host => $host, port => $port, timeout => 600
);
if ($xml) {
  $server->add_method({
    name => "Fulltext.getAnnotation",
    signature => ["struct string"],
    code => \&get_xml_annotation
  });
} else {
  $server->add_method({
    name => "Fulltext.getAnnotation",
    signature => ["struct array"],
    code => \&get_fulltext_annotation
  });
}
$server->add_method({
  name => "Medline.getAnnotation",
  signature => ["struct int"],
  code => \&get_medline_annotation
});
print "starting server at $host:$port...\n";
print "methods:\n";
map { print "\t$_\n";} $server->list_methods();
$server->server_loop();
