#!/usr/bin/python
# encoding: utf-8

"""bc-annotation-server.py

BioCrative Annotation Server (BCAS)

This script is an example of how to implement the BCAS server. The actual
text mining API is implemented in the classes Fulltext and Medline, but
should either be finetuned by your needs or exchanged for your own classes.

Fulltext handles annotation request for full-text articles (as in the BC II.5
challenge), while Medline handles annotation requests for PubMed abstracts (as
in the BioCreative MetaServer prototype) using the PMID as parameter. For the
challenge, only the full-text implementation is necessary.

The XML-RPC API is very simple:
<Type>.getAnnotation
where Type is either 'Fulltext' or 'Medline'.

The client will call this server using either fulltext or the PMID as input
and expects the Annotation Server to return a struct (dict) of result types
("interaction", "normalizations", "pairs", plus three more for the full API
not used in the contest), where each value is an array of structs with the
specific result. These result structs always contain a "confidence" and "rank"
key, except for the array of length one in "interction", where it makes no
sense to report a rank. For details, see the description on the team webpage.
"""

import os
import sys
import xmlrpclib

from optparse import OptionParser
from SimpleXMLRPCServer import SimpleXMLRPCServer

__author__ = "Florian Leitner"
__version__ = "1.0"

DEFAULT_HOST = 'localhost'
DEFAULT_PORT = 8000

def sample_result():
    """Return a structurally valid sample result for all annotation types."""
    return {
        "pairs": [
            {
                "accession_a": "P1234", "accession_b": "Q1234",
                "rank": 1, "confidence": 1.0
            },
            {
                "accession_a": "P1235", "accession_b": "Q1235",
                "rank": 2, "confidence": 0.5
            },
            {
                "accession_a": "P1236", "accession_b": "Q1236",
                "rank": 3, "confidence": 0.0001
            },
        ],
        "interaction": [
            {"has_interaction": 0, "confidence": 0.34},
        ],
        "normalizations": [
            {"accession": "P1234", "rank": 1, "confidence": 1.0},
            {"accession": "P1235", "rank": 2, "confidence": 0.5},
            {"accession": "P1236", "rank": 3, "confidence": 0.0001},
        ],
    }

class Fulltext(object):
    """Sample full-text API for an Annotation Server."""
    
    @staticmethod
    def getAnnotation(text):
        if not isinstance(text, list):
            raise xmlrpclib.Fault(3, 'illegal type for text')
        
        for d in text:
            if not isinstance(d, dict):
                raise xmlrpclib.Fault(
                    3, 'illegal type for text array element'
                )
            
            for k in d:
                if k == "number":
                    if not isinstance(d[k], int):
                        raise xmlrpclib.Fault(
                            3, "text number association not an integer"
                        )
                    continue
                
                if isinstance(d[k], str):
                    d[k] = unicode(d[k], 'utf-8')
                
                if not isinstance(d[k], unicode):
                    raise xmlrpclib.Fault(
                        3, "illegal type for text struct value"
                    )
        
        # call annotation pipeline or command line here
        # for now, just produce the sample result:
        result = sample_result()
        
        return result

class Medline(object):
    """Sample PubMed API for an Annotation Server."""
    
    @staticmethod
    def getAnnotation(pmid):
        try:
            pubmed_id = int(pmid)
        except TypeError:
            raise xmlrpclib.Fault(3, 'illegal type for pmid')
        except ValueError:
            raise xmlrpclib.Fault(4, 'illegal value for pmid')
        
        # here could be a possible place to test the pubmed_id exists:
        #if pubmed_id not in some_set_of_pmids:
        #    raise xmlrpclib.Fault(7, 'no such pmid')
        
        # call annotation pipeline or command line here
        # for now, just produce the sample result:
        result = sample_result()
        
        return result

class XML(object):
    """Sample XML-formated article full-text API for an Annotation Server."""
    
    @staticmethod
    def getAnnotation(text):
        if isinstance(text, str):
            xml_article = unicode(text, 'utf-8')
        elif not isinstance(text, unicode):
            raise xmlrpclib.Fault(
                3,
                "illegal type (%s) for XML text" % str(type(text))
            )
        else:
            xml_article = text
        
        # call annotation pipeline or command line here
        # for now, just produce the sample result:
        result = sample_result()
        
        return result

def run(fulltext_api, medline_api, host, port):
    """Start the server with the two APIs.
    
    In the context of testing the BCAS for the BC II.5 challenge, it does
    not matter if your Medline API actually is working or if you use the
    dummy API provided here.
    """
    server = SimpleXMLRPCServer((host, port))
    print "Listening on %s:%i..." % (host, port)
    server.register_function(
        fulltext_api.getAnnotation, 'Fulltext.getAnnotation'
    )
    server.register_function(
        medline_api.getAnnotation, 'Medline.getAnnotation'
    )
    
    server.serve_forever()

def main():
    usage = "usage: %prog [--options] [[host] port]"
    description = "run the BioCreative Annotation Server"
    epilog = "defaults: host=%s; port=%i" % (DEFAULT_HOST, DEFAULT_PORT)
    parser = OptionParser(
        usage=usage, version=__version__, prog=os.path.basename(sys.argv[0]),
        description=description, epilog=epilog, add_help_option=True
    )
    parser.add_option(
        "-x", "--xml", action="store_true", default=False,
        help="expect XML, not structured UTF-8 from fulltext"
    )
    
    opts, args = parser.parse_args()
    
    l_args = len(args)
    if l_args > 2:
        parser.error("wrong number of arguments (%i/2)" % len(args))
    
    host = DEFAULT_HOST
    port = DEFAULT_PORT
    
    if l_args >= 1:
        if l_args == 2:
            host = args[0]
            port = int(args[1])
        else:
            port = int(args[0])
    
    run(XML if opts.xml else Fulltext, Medline, host, port)

if __name__ == "__main__":
    main()
